import sys 
sys.path.append('.')
from cetl import DataPipeline

cfg = { "pipeline": [   {"type":"dummyStart", "module_type":"functional"},
                        {"type":"parallelTransformer", "transformers":[
                            {"breakpoint":1, "type":"generateDataFrame"},
                            {"type":"generateDataFrame"},
                            {"type":"generateDataFrame"}
                        ]},
                        {"type":"unionAll"}],
                    
        "pipeline_settings":{   "print_cfg":1, 
                                "print_task_result":1, 
                                "exchange_media":"default"}}

pipe = DataPipeline(cfg)
for step in pipe.steps:
    task_id, transformer = step
    print(transformer)
df = pipe.transform("")
# print(df)


# pipe = pipe.build_digraph()
# pipe.save_png("./sample7.png")


# python cetl/dev_tests/sample9_build_pipeline_from_cfg2.py


# pip install kafka-python
# pip install avro-python3
from kafka import KafkaProducer, KafkaConsumer, TopicPartition
from kafka.admin import KafkaAdminClient, NewTopic
import json
import pandas as pd
from kafka import KafkaProducer
import io
import pandas as pd
import pickle
from cryptography.fernet import Fernet
from kafka.errors import UnknownTopicOrPartitionError

class kafkaHelper:
    """usage
    # Define Kafka connection parameters
    bootstrap_servers = ['192.168.32.1:9092']
    topic_name = 'mytopic2'
    # task_id
    task_id = b"1.filterBy"
    # Create a pandas dataframe
    df = pd.DataFrame({'col1': [1, 2, 3], 'col2': ['a', 'b', 'c']})

    with kafkaHelper(   bootstrap_servers=bootstrap_servers,
                        topic=topic_name) as kh:

        # serialize the dataframe to pickle
        serialized_data = km.serialize_dataframe(df)
        # print("serialized_data: ", serialized_data)

        # create instance of encryption
        encrypted_data = kh.fernet.encrypt(serialized_data)
        # print(encrypted_data)

        # send the data to kafka topic
        km.send(key=task_id, value=encrypted_data)


        # received message
        message = kh.receive(task_id)

        # decrypt the message
        decrypted_data = kh.fernet.decrypt(message)

        # decode by pickle
        df = kh.deserialize_dataframe(decrypted_data)

        print(df)

        # delete the topic
        kh.delete_topic()
    """
    def __init__(self,
                bootstrap_servers=None,
                topic=None):

        self.bootstrap_servers = bootstrap_servers
        self.topic = topic
        self.producer=None
        self.consumer=None
        

    def __enter__(self,):
        
        try:
            # self.admin_client = KafkaAdminClient(bootstrap_servers=self.bootstrap_servers)
            # try:
            #     self.delete_topic()
            # except Exception as e:
            #     print("delete_topic got error")

            # Create a Kafka producer instance
            # self.producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers)

            # create instance of KafkaConsumer
            self.consumer = KafkaConsumer(self.topic,
                                    bootstrap_servers=self.bootstrap_servers,
                                    auto_offset_reset='earliest',
                                    enable_auto_commit=True)
        except Exception as e:
            # print("####################", e)
            # print("close producer instance")
            # self.producer.close()
            # self.consumer.close()
            print("has error")
        
        return self


    def __exit__(self, exc_type, exc_value, exc_tb):
        print("test when will run exit ################################")
        self.producer.close()
        self.consumer.close()

    def topic_exists(self, admin_client, topic_name):
        """
        [{'error_code': 3, 'topic': 'c', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 'o', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': '3', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 'm', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 't', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 'p', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 'y', 'is_internal': False, 'partitions': []}, 
        {'error_code': 3, 'topic': 'i', 'is_internal': False, 'partitions': []}]
        """
        try:
            topic_metadata = admin_client.describe_topics([topic_name])
            print(topic_metadata)
            # for partition_data in topic_metadata:
            #     if "error_code" in partition_data:

            # print("show topic_metadata", topic_metadata)
            if "error_code" in topic_metadata[0]:
                if topic_metadata[0]["error_code"]==3:
                    return False
            
            return True
        except UnknownTopicOrPartitionError:
            return False


    def delete_topic(self):   
        if not self.admin_client:
            # call back
            print("call back #####")
            self.__enter__()
        

        if self.topic_exists(self.admin_client, self.topic):
            print("deleting###########################", self.topic)
            self.admin_client.delete_topics([self.topic])


    def send(self, key=None, value=None):
        # if not self.producer:
            # print("self.producer is none ##############")
            # self.__enter__()

        print(f"open topic: {self.topic}")
        print(f"key: {key}")
        print(f"value: {value}")
        print("self.producer", self.producer)
        print("bootstrap_servers", self.bootstrap_servers)
        self.producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers)
        self.producer.send(self.topic, key="1.dummyStart", value=value)
        # try:
        #     self.producer.send(self.topic, key=b'1.dummyStart', value=value)
        # except Exception as e:
        #     print("self.producer.send() got error")
        #     print(str(e))
        # producer.flush()


    def receive(self, key):
        if not self.consumer:
            self.__enter__()

        for message in self.consumer:
            print("message.key", message.key)
            print("message.value", message.value)
            if message.key == key:
                received_message = message.value
                self.consumer.close()
                return received_message


    def close(self):
        self.producer.close()
        self.consumer.close()


    def serialize_dataframe(self, df):
        return pickle.dumps(df)

    
    def deserialize_dataframe(self, serialized_data):
        # deserialize the data
        file_obj = io.BytesIO(serialized_data)
        df = pickle.load(file_obj)
        return df

    
# class kafkaMedia(kafkaHelper):
#     def __init__(self, 
#                 bootstrap_servers=None,
#                 topic=None):
#         super().__init__(bootstrap_servers=bootstrap_servers,
#                         topic=topic)


#     def python2kafka(self, key=None, value=None):

#         if isinstance(value, pd.DataFrame):
#             df = value

#             # serialize the dataframe to pickle
#             serialized_data = self.serialize_dataframe(df)
#             # print("serialized_data: ", serialized_data)

#             # create instance of encryption
#             encrypted_data = self.fernet.encrypt(serialized_data)
#             # print(encrypted_data)

#             # send the data to kafka topic
#             self.send(key=key, value=encrypted_data)

#     def kafka2python(self, key=None):

        
#         # received message
#         message = self.receive(key)

#         # decrypt the message
#         decrypted_data = self.fernet.decrypt(message)

#         # decode by pickle
#         df = self.deserialize_dataframe(decrypted_data)

#         # delete the topic
#         # kh.delete_topic()

#         return df





import sys 
sys.path.append('.')
from cetl import DataPipeline


cfg = { "pipeline": [   {"type":"dummyStart"},
                        {"type":"parallelTransformer", "transformers":[
                            {"type":"generateDataFrame"},
                            {"type":"generateDataFrame"},
                            {"type":"generateDataFrame"}
                        ]},
                        {"type":"unionAll"}],
                    
        "pipeline_settings":{   "print_cfg":1, 
                                "print_task_result":1, 
                                "exchange_media":"kafka",
                                "bootstrap_servers":["192.168.176.1:9092"]}}




# pipe = build_cetl_pipeline(cfg, bootstrap_servers=["192.168.176.1:9092"])
pipe = DataPipeline(cfg)
for step in pipe.steps:
    task_id, transformer = step
    print(transformer)
result = pipe.transform("")

# get the output of transformer by task_id
task_id = b"3.generateDataFrame"
df = pipe.kafka_media.read_kafka(task_id=task_id)
print(df)

# delete the topic
pipe.kafka_media.delete_topic()



pipe = pipe.build_digraph()
pipe.save_png("./sample11.png")


# python cetl/dev_tests/sample11_adding_kafka_media.py


from functools import wraps
import time
from .builder import context_name, DataFrame
from datetime import datetime
from .file_mgt import get_datetime_str
from .kafka import kafka2python, python2kafka
EXCH_MEDIA = "default"


def recursive_records_count(count, data):
    if isinstance(data, dict):
        if context_name in data:
            count += len(data[context_name])
        elif "key" in data:
            return count
        else:
            print("not acceptable context_name")
            # raise ValueError()
            return count

    elif isinstance(data, DataFrame):
        count += data.shape[0]
    elif isinstance(data, str):
        count += 0
    elif isinstance(data, list):
        for item in data:
            count = recursive_records_count(count, item)
    else:
        raise ValueError("currently only accept list, str, json and pandas data type")
    
    return count


def transform_wrapper(func):
    """
    Usage
    ----------------------------
    class Calculator:
        @transform_wrapper
        def calculate_something(self, num):
            total = sum((x for x in range(0, num**2)))
            return total

        def __repr__(self):
            return f'calc_object:{id(self)}'
    Reference
    ---------------------------
    https://dev.to/kcdchennai/python-decorator-to-measure-execution-time-54hk
    """

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        transformer = args[0]
        input = args[1]
        kafka_media = transformer.kafka_media


        # Calculate the execution time
        start_time = time.perf_counter()
        transformer.start_datetime = get_datetime_str(format="%Y-%m-%d %H:%M:%S")

        # print processing label
        print(f"Processing {transformer.node_name}, started at {transformer.start_datetime} ...") if transformer.print_task_result else ""

        # execute the transform function -----------------------------------------------------
        if transformer.exchange_media == "default":
            result = func(*args, **kwargs) #result here is module output

        elif transformer.exchange_media =="kafka":
            # decode the input from kafka topic
            pre_module_output=""

            if not transformer.is_first_trans:

                if isinstance(input, dict):
                    pre_transformer_key = input["key"]
                    pre_module_output = kafka_media.read_kafka(   task_id=pre_transformer_key.encode('utf-8'))
                
                elif isinstance(input, list):
                    # previous transformer is wrapped by parallelTransformer
                    pre_module_output = []
                    for pre_transformer_key in [item["key"] for item in input]:
                        # ---------------------------------------------------------- read message
                        kafka2python = kafka_media.read_kafka(task_id=pre_transformer_key.encode('utf-8'))
                        pre_module_output.append(kafka2python)
                    
                else:
                    # print("pre_transformer_key", pre_transformer_key)
                    pre_module_output = kafka_media.read_kafka(   task_id=pre_transformer_key.encode('utf-8'))
                    
            else:
                pre_module_output = ""

            # print("####################")            
            result = func(transformer, pre_module_output) # result here is module output

        else:
            # reserve for future case
            # result = func(*args, **kwargs) #result here is module output
            print(transformer.exchange_media, transformer.node_name)
            assert False

        # ----------------------------------------------------------------------------------------
        # Calcuate the end time
        end_time = time.perf_counter()
        total_time = end_time - start_time
        transformer.end_datetime = get_datetime_str(format="%Y-%m-%d %H:%M:%S")
        # print(f'Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds')


        # Adding the execution time to the transformer
        transformer.total_time = total_time
        
        # print(self)
        # parallelTransformer will make it run double times, not solved yet

        #count the number of records in input and output of the func:
        transformer.total_input = recursive_records_count(0, input)
        transformer.total_output = recursive_records_count(0, result)

        if transformer.node_name:
            print(f"    - take time {total_time:.4f} seconds, total input records {transformer.total_input}, total output records {transformer.total_output}") if transformer.print_task_result else ""

            # print(f"stop with breakpoint at {transformer.node_name}") if transformer.print_task_result else ""
        # print(f"{transformer.node_name}: take time {total_time:.4f} seconds, total input records {transformer.total_input}, total output records {transformer.total_output}")

        # output result  -------------------------------------------------------------------------
        if transformer.exchange_media == "default":
            return result

        elif transformer.exchange_media =="kafka":
            # print(transformer)
            # print("####################")
            # sending the python object to kafka topic
            # print("transformer.node_name", transformer.node_name.encode('utf-8'))
            # print("result", result)
            kafka_media.send_kafka(task_id=transformer.node_name.encode('utf-8'), value=result)

            result = {  "key":transformer.node_name,
                        "total_input":transformer.total_input, 
                        "total_output":transformer.total_output}
            return result
        else:
            # reserve for future use
            # return result

            print(transformer.exchange_media, transformer.node_name)
            assert False
        # ----------------------------------------------------------------------------------------

    return func_wrapper


    

    





# pip install kafka-python
# pip install avro-python3
from kafka import KafkaProducer, KafkaConsumer, TopicPartition
from kafka.admin import KafkaAdminClient, NewTopic
from kafka.errors import UnknownTopicOrPartitionError
import json
import pandas as pd
from kafka import KafkaProducer
from avro import schema, io
import io
import pandas as pd
import pickle
from cryptography.fernet import Fernet



class kafkaMedia:
    def __init__(self,
                bootstrap_servers=None,
                topic=None):

        self.bootstrap_servers = bootstrap_servers
        self.topic = topic
        self.producer=None
        self.consumer=None
        # print("init fernet ##########################")
        self.pub_key = Fernet.generate_key()
        self.fernet = Fernet(self.pub_key)
        self.admin_client = KafkaAdminClient(bootstrap_servers=self.bootstrap_servers)


    def __enter__(self,):
        
        self.delete_topic()

        # Create a Kafka producer instance
        self.producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers)

        # create instance of KafkaConsumer
        self.consumer = KafkaConsumer(self.topic,
                                bootstrap_servers=self.bootstrap_servers,
                                auto_offset_reset='earliest',
                                enable_auto_commit=True)

        try:
            self.admin_client = KafkaAdminClient(bootstrap_servers=self.bootstrap_servers)
            self.delete_topic()

            # Create a Kafka producer instance
            self.producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers)

            self.consumer = KafkaConsumer(self.topic,
                                    bootstrap_servers=self.bootstrap_servers,
                                    auto_offset_reset='earliest',
                                    enable_auto_commit=True)

        except Exception as e:
            # print("####################", e)
            # print("close producer instance")
            # if self.producer:
            #     self.producer.close()
            # if self.consumer:
            #     self.consumer.close()
            print("has error")
            pass
        
        return self


    def __exit__(self, exc_type, exc_value, exc_tb):
        print("# test when will run exit ####################")
        if self.producer:
            self.producer.close()
        if self.consumer:
            self.consumer.close()

    def topic_exists(self, admin_client, topic_name):
        """
        """
        try:
            topic_metadata = admin_client.describe_topics([topic_name])
            # for partition_data in topic_metadata:
            #     if "error_code" in partition_data:

            # print("show topic_metadata", topic_metadata)
            if "error_code" in topic_metadata[0]:
                if topic_metadata[0]["error_code"]==3:
                    return False

                if topic_metadata[0]["error_code"]==0:
                    return True
            
            return True
        except UnknownTopicOrPartitionError:
            return False


    def delete_topic(self):
        # if not self.admin_client:
            # self.__enter__()
        if self.topic_exists(self.admin_client, self.topic):
            print("delete topic ...")
            self.admin_client.delete_topics([self.topic])


    def send(self, key=None, value=None):
        # if not self.producer:
            # self.__enter__()
        # print("topic", self.topic)
        # print("send key", key)
        self.producer.send(self.topic, key=key, value=value)
        # producer.flush()

    def send_kafka(self, task_id=None, value=None):

        df = value

        # serialize the dataframe to pickle
        serialized_data = self.serialize_dataframe(df)
        # print("serialized_data: ", serialized_data)

        # create instance of encryption
        encrypted_data = self.fernet.encrypt(serialized_data)
        # print(encrypted_data)

        # send the data to kafka topic
        # print("send key", task_id)
        self.send(key=task_id, value=encrypted_data)
        # print("send encrypted_data", encrypted_data)


    def receive(self, key=None):
        # if not self.consumer:
        #     # create instance of KafkaConsumer
        #     self.consumer = KafkaConsumer(self.topic,
        #                             bootstrap_servers=self.bootstrap_servers,
        #                             auto_offset_reset='earliest',
        #                             enable_auto_commit=True)

        # print("self.consumer", self.consumer)
        for message in self.consumer:
            # print("dectect message", message.key, key)
            # print(message.key, key)
            if message.key == key:
                received_message = message.value
                return received_message


    def receive_with_partition(self, key=None):
        if not self.consumer:
            # create instance of KafkaConsumer
            self.consumer = KafkaConsumer(self.topic,
                                    bootstrap_servers=self.bootstrap_servers,
                                    auto_offset_reset='earliest',
                                    enable_auto_commit=True)

        # Get partitions for the topic
        partitions = self.consumer.partitions_for_topic(self.topic)

        # Loop through partitions and clear messages
        for partition in partitions:
            # Get current partition offset
            tp = TopicPartition(self.topic, partition)
            current_offset = self.consumer.position(tp)

            # Seek to beginning of partition
            self.consumer.seek_to_beginning(tp)

            # Consume messages and send null messages with same keys to clear the topic
            for message in self.consumer:
                # print("detect message", message.key, key)
                if message.key == key:
                    received_message = message.value
                    return received_message
            
            

    def read_kafka(self, task_id=None):
        # received message
        # print("self.receive###############", self.receive)
        message = self.receive_with_partition(key=task_id)
        assert message
        # print("received message", message)
        
        # decrypt the message
        # print(key)
        decrypted_data = self.fernet.decrypt(message)

        # # decode by pickle
        df = self.deserialize_dataframe(decrypted_data)

        return df


    def close(self):
        self.producer.close()
        self.consumer.close()


    def serialize_dataframe(self, df):
        return pickle.dumps(df)

    
    def deserialize_dataframe(self, serialized_data):
        # deserialize the data
        file_obj = io.BytesIO(serialized_data)
        df = pickle.load(file_obj)
        return df


# from .kafka_helper import kafkaHelper
# from ..utils.builder import pd
# from cryptography.fernet import Fernet

# class kafkaMedia(kafkaHelper):
#     def __init__(self, 
#                 bootstrap_servers=None,
#                 topic=None):
#         super().__init__(bootstrap_servers=bootstrap_servers,
#                         topic=topic)

#     def send_kafka(self, key=None, value=None):

#         if isinstance(value, pd.DataFrame):
#             df = value

#             # serialize the dataframe to pickle
#             serialized_data = self.serialize_dataframe(df)
#             # print("serialized_data: ", serialized_data)

#             # create instance of encryption
#             self.pub_key = Fernet.generate_key()
#             self.fernet = Fernet(self.pub_key)

#             encrypted_data = self.fernet.encrypt(serialized_data)
#             # print("encrypted_data", encrypted_data)

#             # send the data to kafka topic
#             self.send(key=key, value=encrypted_data)

#     def read_kafka(self, key=None):

        
#         # received message
#         message = self.receive(key=key)

#         # decrypt the message
#         decrypted_data = self.fernet.decrypt(message)

#         # decode by pickle
#         df = self.deserialize_dataframe(decrypted_data)

#         # delete the topic
#         # kh.delete_topic()

        return df


